; Game Raccoon Revision 0 Source Code
; Copyright (C) 2019 Mathew Carr

; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU Affero General Public License Version 3
; with Additional Terms as found in the file
; LICENSE_AGPL_V3_ADDITIONAL_TERMS.txt.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU Affero General Public License for more details.

; You should have received a copy of the GNU Affero General Public License
; along with this program.  If not, see <https://www.gnu.org/licenses/>.
    
;;

; This file is the 'outer layer' of the Game Raccoon boot image.
; It exists to provide the Mega Drive header, and a tiny bootstrap
; header that copies the Game Raccoon RAM-resident code into the
; Mega Drive RAM and jumps to it.

        include "megadrive_defines.68k.asm"

STACK_POINTER_START = RAM_END    ; Full descending stack points to address beyond end of RAM.

CART_ENTRY_POINT    = main

; Game Raccoon does not support interrupts of any kind.
unhandled_exception = $00000000
vblank              = $00000000
hblank              = $00000000

VECTOR_BUS_ERROR                            = unhandled_exception
VECTOR_ADDRESS_ERROR                        = unhandled_exception
VECTOR_ILLEGAL_INSTRUCTION                  = unhandled_exception
VECTOR_DIVISION_BY_ZERO                     = unhandled_exception
VECTOR_CHK_EXCEPTION                        = unhandled_exception
VECTOR_TRAPV_EXCEPTION                      = unhandled_exception
VECTOR_PRIVILEGE_VIOLATION_EXCEPTION        = unhandled_exception
VECTOR_TRACE_EXCEPTION                      = unhandled_exception
VECTOR_LINE_1010_EMULATOR                   = unhandled_exception
VECTOR_LINE_1111_EMULATOR                   = unhandled_exception
VECTOR_RESERVED_EXCEPTION                   = unhandled_exception
VECTOR_SPURIOUS_EXCEPTION                   = unhandled_exception
VECTOR_INTERRUPT_REQUEST_LEVEL_1            = unhandled_exception
VECTOR_INTERRUPT_REQUEST_LEVEL_2            = unhandled_exception
VECTOR_INTERRUPT_REQUEST_LEVEL_3            = unhandled_exception
VECTOR_INTERRUPT_REQUEST_LEVEL_4_VDP_HBLANK = hblank
VECTOR_INTERRUPT_REQUEST_LEVEL_5            = unhandled_exception
VECTOR_INTERRUPT_REQUEST_LEVEL_6_VDP_VBLANK = vblank
VECTOR_INTERRUPT_REQUEST_LEVEL_7            = unhandled_exception
VECTOR_TRAP_00_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_01_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_02_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_03_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_04_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_05_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_06_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_07_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_08_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_09_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_10_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_11_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_12_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_13_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_14_EXCEPTION                    = unhandled_exception
VECTOR_TRAP_15_EXCEPTION                    = unhandled_exception

        ; ROM Header.
        
        ; Bytes $08 to $FF are the Motorola 68000 exception vector table.

        org $00
        dc.l STACK_POINTER_START
        dc.l CART_ENTRY_POINT
        dc.l VECTOR_BUS_ERROR
        dc.l VECTOR_ADDRESS_ERROR

        dc.l VECTOR_ILLEGAL_INSTRUCTION
        dc.l VECTOR_DIVISION_BY_ZERO
        dc.l VECTOR_CHK_EXCEPTION
        dc.l VECTOR_TRAPV_EXCEPTION

        dc.l VECTOR_PRIVILEGE_VIOLATION_EXCEPTION
        dc.l VECTOR_TRACE_EXCEPTION
        dc.l VECTOR_LINE_1010_EMULATOR
        dc.l VECTOR_LINE_1111_EMULATOR

        dc.l unhandled_exception                              ; Reserved interrupt vector entries that aren't implemented in 68000.
        dc.l unhandled_exception
        dc.l unhandled_exception
        dc.l unhandled_exception

        dc.l unhandled_exception
        dc.l unhandled_exception
        dc.l unhandled_exception
        dc.l unhandled_exception

        dc.l unhandled_exception
        dc.l unhandled_exception
        dc.l unhandled_exception
        dc.l unhandled_exception

        dc.l VECTOR_SPURIOUS_EXCEPTION
        dc.l VECTOR_INTERRUPT_REQUEST_LEVEL_1
        dc.l VECTOR_INTERRUPT_REQUEST_LEVEL_2
        dc.l VECTOR_INTERRUPT_REQUEST_LEVEL_3

        dc.l VECTOR_INTERRUPT_REQUEST_LEVEL_4_VDP_HBLANK
        dc.l VECTOR_INTERRUPT_REQUEST_LEVEL_5
        dc.l VECTOR_INTERRUPT_REQUEST_LEVEL_6_VDP_VBLANK
        dc.l VECTOR_INTERRUPT_REQUEST_LEVEL_7

        dc.l VECTOR_TRAP_00_EXCEPTION
        dc.l VECTOR_TRAP_01_EXCEPTION
        dc.l VECTOR_TRAP_02_EXCEPTION
        dc.l VECTOR_TRAP_03_EXCEPTION

        dc.l VECTOR_TRAP_04_EXCEPTION
        dc.l VECTOR_TRAP_05_EXCEPTION
        dc.l VECTOR_TRAP_06_EXCEPTION
        dc.l VECTOR_TRAP_07_EXCEPTION

        dc.l VECTOR_TRAP_08_EXCEPTION
        dc.l VECTOR_TRAP_09_EXCEPTION
        dc.l VECTOR_TRAP_10_EXCEPTION
        dc.l VECTOR_TRAP_11_EXCEPTION

        dc.l VECTOR_TRAP_12_EXCEPTION
        dc.l VECTOR_TRAP_13_EXCEPTION
        dc.l VECTOR_TRAP_14_EXCEPTION
        dc.l VECTOR_TRAP_15_EXCEPTION
        
        ; All other exceptions are FPU or MMU related and unhandled.

        org $100
        dc.b "SEGA MEGA DRIVE "                    ; 16 - Console name
        org $110
        dc.b "(C)MATT 2020.JAN"                    ; 16 - Release date "(C)comp year.mon" (TODO: Update this appropriately?)
        org $120
        dc.b "GAME RACCOON BOOT IMAGE MATHEW CARR" ; 48 - Domestic name
        org $150
        dc.b "GAME RACCOON BOOT IMAGE MATHEW CARR" ; 48 - International name
        org $180
        dc.b "GR GRACCOON-00"                      ; 14 - Version string
        org $18E
        dc.w $1234                                 ;  2 - Checksum (to be done later)
        org $190
        dc.b "J               "                    ; 16 - Device support. J is joypad. Don't support lightguns, modems, etc.
        org $1A0
        dc.l $00000000                             ;  4 - ROM start
        dc.l $0000FFFF                             ;  4 - ROM end (TODO: Magically change this to be ROM size, also padded to nearest 64kbyte)
        org $1A8
        dc.l RAM_START                             ;  4 - RAM start
        dc.l RAM_LAST_BYTE                         ;  4 - RAM end
        org $1B0
        dc.b "            "                        ; 12 - Non-volatile memory information area
        org $1BC
        dc.b "            "                        ; 12 - Modem support
        org $1C8
        dc.b "          "                          ; 40 - Reserved
        dc.b "          "
        dc.b "          "
        dc.b "          "
        org $1F0
        dc.b "JUE"                                 ;  3 - Region support.

; Header ends.

; RAM memory map assignments for Game Raccoon boot ROM:

; The RAM-resident Game Raccoon executable image is installed
; at the front of RAM.
RACCOON_RAM_RESIDENT_DESTINATION = RAM_START

        org $200
; We're going to put the binaries blob at $200 so we can calculate all the offsets
; and lengths in advance. 'main' will be pushed down into the ROM by the binaries
; archive, but that is of no consequence since the Mega Drive cart header is
; generated from the 'main' symbol.

        incbin  "included_binaries.bin"

        align   2
; Entry point when cart is booted.
main:
        ; Initialise Mega Drive:
        ; We want to satisfy the TMSS check as soon as possible.
        interrupts_disable
        move.b  (MM_VERSION_REGISTER+1),d0       ; Load the low byte from the version register.
        and.b   #$0F,d0
        beq     tmss_satisfy_tmss_unnecessary
        move.l  #"SEGA",(MM_TMSS_REGISTER)       ; Satisy TMSS on version != zero consoles by writing SEGA longword to its register.
tmss_satisfy_tmss_unnecessary:                   ; Branch here if we're on a version zero console that does not require TMSS compliance.
        
        ; Past this point, the console is entirely at our command.
        ; I'm going to copy the Game Raccoon Executable Area
        ; into RAM $FF0000...
        lea     raccoon_ram_resident_rom_included_block_start,a0
        lea     RACCOON_RAM_RESIDENT_DESTINATION,a1
        move.w  #(RACCOON_RAM_RESIDENT_ROM_INCLUDED_BLOCK_LENGTH_LONGS-1),d0
copy_next_long:        
        move.l  (a0)+,(a1)+
        dbf     d0,copy_next_long

        jmp     RACCOON_RAM_RESIDENT_DESTINATION
        
        ; Include the RAM-resident portion of the code contiguous in ROM,
        ; but with instructions based at the corresponding position in RAM.
        align   2        

raccoon_ram_resident_rom_included_block_start:        
        ; This file must always be a multiple of 4 large.
        ; (All the variables it messes with expect to begin on a long aligned address in RAM.)
        ; I want to ensure that in the assembly of raccoonramresident.68k.asm.
        incbin  "raccoonramresident.bin"
raccoon_ram_resident_rom_included_block_end = *

RACCOON_RAM_RESIDENT_ROM_INCLUDED_BLOCK_LENGTH_BYTES = raccoon_ram_resident_rom_included_block_end-raccoon_ram_resident_rom_included_block_start
RACCOON_RAM_RESIDENT_ROM_INCLUDED_BLOCK_LENGTH_LONGS = (RACCOON_RAM_RESIDENT_ROM_INCLUDED_BLOCK_LENGTH_BYTES/4)