# Game Raccoon Revision 0 Source Code
# Copyright (C) 2019 Mathew Carr

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License Version 3
# with Additional Terms as found in the file
# LICENSE_AGPL_V3_ADDITIONAL_TERMS.txt.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
    
##

# This script reads a local folder specified below
# and outputs a .bin file that has the same format
# as that generated by the Game Raccoon PIC
# directory cache generation command.
# You can incbin this into a GameRaccoon boot rom
# at $200000 (bytes address) to test the boot rom
# in an emulator (since there's no physical or 
# emulated Game Raccoon to communicate with to
# provide it).
# This will be overwritten the next time
# the Game Raccoon regenerates the directory cache
# itself.
import glob
import os
import json
import random 

def get_directory_tree_populate(d):
   # Given a directory element with the 'name' key
   # populated, fill its fields.
   
   for x in glob.glob(d["name"] + "/*"):
      if os.path.isfile(x):
         d["files"].append(x)
   
      if os.path.isdir(x):
         newdir = {"name": x, "directories": [], "files": []}
         d["directories"].append(newdir)
         get_directory_tree_populate(newdir)
   
   random.shuffle(d["files"])
   random.shuffle(d["directories"])
   
   pass
   
def get_directory_tree(path):
   # With path as the . directory,
   # make some kind of useful structure we can
   # use to imitate the layout of a Game Raccoon
   # directory cache.
   oldwdir = os.getcwd()
   
   os.chdir(path)
   
   root_dir = {"name": ".", "directories": [], "files": []}
   
   get_directory_tree_populate(root_dir)
   
   os.chdir(oldwdir)
   
   return root_dir
   
def value_to_big_endian_byte_pair(v):
   lsb = (v & 0x00FF) >> 0;
   msb = (v & 0xFF00) >> 8;
   
   return bytes([msb, lsb])
   
def directory_tree_to_game_raccoon_cache_binary_do_directory(element, index_of_self_in_output_list, workspace_object):
   my_contents_index = workspace_object["writing_cursor"] + 1

   workspace_object["output_list"][index_of_self_in_output_list]["contents_start"] = my_contents_index
   workspace_object["output_list"][index_of_self_in_output_list]["name"]           = element["name"]
   workspace_object["output_list"][index_of_self_in_output_list]["no_directories"] = len(element["directories"])
   workspace_object["output_list"][index_of_self_in_output_list]["no_files"]       = len(element["files"])
   
   for d in element["directories"]:
      new_dir_output_line = {"type": "DIR", "parent": index_of_self_in_output_list}
      workspace_object["writing_cursor"] += 1
      workspace_object["output_list"].append(new_dir_output_line)
   
   for f in element["files"]:
      new_file_output_line = {"type": "FILE", "name": f}
      workspace_object["writing_cursor"] += 1
      workspace_object["output_list"].append(new_file_output_line)
   
   for i,d in enumerate(element["directories"]):
      directory_tree_to_game_raccoon_cache_binary_do_directory(d, my_contents_index+i, workspace_object)
      
def directory_tree_to_game_raccoon_cache_binary(output_filename, root_element):
   #print(json.dumps(root_element, indent=3))

   with open(output_filename, "wb") as o:
      workspace_object = {"writing_cursor": 0, "output_list": []}
      
      # Given that the directory item for this directory
      # exists, output the contents.
      
      workspace_object["output_list"].append({
         "type":           "DIR",
         "parent":         0
      })
      
      directory_tree_to_game_raccoon_cache_binary_do_directory(root_element, 0, workspace_object)
      
      #print(json.dumps(workspace_object["output_list"], indent=3))
      #exit(0)
      
      for item in workspace_object["output_list"]:
         if item["type"] == "DIR":
            asciiname = item["name"].encode("ascii")
            asciiname = asciiname.replace(b'\\', b'/')
            asciiname_padded = asciiname + b'\x00' * ((256-(4*2))-len(asciiname))
            
            o.write(asciiname_padded +
                    value_to_big_endian_byte_pair(item["no_directories"]) +
                    value_to_big_endian_byte_pair(item["no_files"]) +
                    value_to_big_endian_byte_pair(item["contents_start"]) +
                    value_to_big_endian_byte_pair(item["parent"]))
         
         if item["type"] == "FILE":
            asciiname = item["name"].encode("ascii")
            asciiname = asciiname.replace(b'\\', b'/')
            asciiname_padded = asciiname + b'\x00' * ((256-(0*2))-len(asciiname))
            
            o.write(asciiname_padded)
     
directory_tree_to_game_raccoon_cache_binary("directory_cache_emulator.bin", get_directory_tree('C:\\Emulators\\roms\\megadrive\\GameRaccoon_folders'))