; Game Raccoon Revision 0 Source Code
; Copyright (C) 2019 Mathew Carr

; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU Affero General Public License Version 3
; with Additional Terms as found in the file
; LICENSE_AGPL_V3_ADDITIONAL_TERMS.txt.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU Affero General Public License for more details.

; You should have received a copy of the GNU Affero General Public License
; along with this program.  If not, see <https://www.gnu.org/licenses/>.
    
;;

RAM_START                            = $00FF0000
RAM_END                              = $01000000
RAM_LAST_BYTE                        = (RAM_END-1)
RAM_SIZE_BYTES                       = $00010000
RAM_SIZE_WORDS                       = RAM_SIZE_BYTES/2
RAM_SIZE_LONGS                       = RAM_SIZE_BYTES/4

VRAM_SIZE_BYTES                      = $00010000
VRAM_SIZE_WORDS                      = VRAM_SIZE_BYTES/2
VRAM_SIZE_LONGS                      = VRAM_SIZE_BYTES/4

VSRAM_SIZE_WORDS                     = 40
VSRAM_SIZE_BYTES                     = VSRAM_SIZE_WORDS*2
VSRAM_SIZE_LONGS                     = VSRAM_SIZE_BYTES/4

CRAM_SIZE_WORDS                      = 16*4
CRAM_SIZE_BYTES                      = CRAM_SIZE_WORDS*2

PATTERN_SIZE_BYTES                   = ((8*8*4)/8)            ; A pattern is an 8x8 tile, 4 bits per pixel depth.
PATTERN_SIZE_WORDS                   = PATTERN_SIZE_BYTES/2

;;;

; $000.l = address to set stack on game boot. MUST BE EVEN!
; $100.l = "SEGA"
MM_STACK_POSITION                    = $000
MM_ENTRY_POINT                       = $004
MM_SEGA_TEXT_HEADER                  = $100

; Equates for the Mega Drive's memory mapped peripherals.
MM_Z80_MEMORY_SPACE                  = $00A00000

MM_VERSION_REGISTER                  = $00A10000 ; Read only word long.

MM_CONTROLLER_1_DATA                 = $00A10002
MM_CONTROLLER_2_DATA                 = $00A10004
MM_EXPANSION_PORT_DATA               = $00A10006

MM_CONTROLLER_1_CONTROL              = $00A10008
MM_CONTROLLER_2_CONTROL              = $00A1000A
MM_EXPANSION_PORT_CONTROL            = $00A1000C

; Use this when we're cycling through the pads and we want to nn(An) access the CONTROL part
; when An points to the data register.
MM_CONTROLLER_DATA_TO_CONTROL_OFFSET = MM_CONTROLLER_1_CONTROL-MM_CONTROLLER_1_DATA
MM_CONTROLLER_DATA_SOURCE_STRIDE     = MM_CONTROLLER_2_DATA-MM_CONTROLLER_1_DATA

MM_CONTROLLER_1_SERIAL_TX            = $00A1000E
MM_CONTROLLER_1_SERIAL_RX            = $00A10010
MM_CONTROLLER_1_SERIAL_CONTROL       = $00A10012
MM_CONTROLLER_2_SERIAL_TX            = $00A10014
MM_CONTROLLER_2_SERIAL_RX            = $00A10016
MM_CONTROLLER_2_SERIAL_CONTROL       = $00A10018
MM_EXPANSION_PORT_SERIAL_TX          = $00A1001A
MM_EXPANSION_PORT_SERIAL_RX          = $00A1001C
MM_EXPANSION_PORT_SERIAL_CONTROL     = $00A1001E

MM_MEMORY_MODE_REGISTER              = $00A11000

MM_Z80_BUS_REQUEST                   = $00A11100
MM_Z80_RESET                         = $00A11200

MM_TIME_ADDRESS_BEGIN                = $00A13000 ; Writes to this area will cause !TIME to be pulled low in cart, this hits the Game Raccoon TIME register.
MM_TIME_ADDRESS_END                  = $00A13100

MM_TMSS_REGISTER                     = $00A14000 ; Longword sized.

MM_VDP_DATA                          = $00C00000
MM_VDP_CONTROL                       = $00C00004
MM_VDP_HV_COUNTER                    = $00C00008

MM_SN76489_PSG                       = $00C00011 ; Write bytes only.

MM_YM2612_ADDRESS_P1                 = $00A04000 ; Write bytes only.
MM_YM2612_DATA_P1                    = $00A04001
MM_YM2612_ADDRESS_P2                 = $00A04002
MM_YM2612_DATA_P2                    = $00A04003

YM2612_STATUS_BYTE_BUSY_BIT          = 7

YM2612_22_LFO_FREQ                   = $22
YM2612_24_TIMER_A_MSB                = $24
YM2612_25_TIMER_A_LSB                = $25
YM2612_26_TIMER_B_MSB                = $26
YM2612_27_CH3_MODE_TIMER_ENABLES     = $27
YM2612_28_NOTE_ON_CONTROL            = $28
YM2612_2A_DAC_VALUE                  = $2A
YM2612_2B_DAC_ENABLE                 = $2B

; Constants with the write bit already set.
VDP_REGISTER_WRITE_COMMAND                                   = $8000
VDP_REGISTER_MODE_REGISTER_1                                 = VDP_REGISTER_WRITE_COMMAND|$0004 ; We're going to cheat and put the 1 that's expected here.
VDP_REGISTER_MODE_REGISTER_2                                 = VDP_REGISTER_WRITE_COMMAND|$0100
VDP_REGISTER_PLANE_A_NAME_TABLE_LOCATION                     = VDP_REGISTER_WRITE_COMMAND|$0200
VDP_REGISTER_WINDOW_NAME_TABLE_LOCATION                      = VDP_REGISTER_WRITE_COMMAND|$0300
VDP_REGISTER_PLANE_B_NAME_TABLE_LOCATION                     = VDP_REGISTER_WRITE_COMMAND|$0400
VDP_REGISTER_SPRITE_TABLE_LOCATION                           = VDP_REGISTER_WRITE_COMMAND|$0500
VDP_REGISTER_UNUSED_SPRITE_PATTERN_GENERATOR_BASE_ADDRESS    = VDP_REGISTER_WRITE_COMMAND|$0600
VDP_REGISTER_BACKGROUND_COLOUR                               = VDP_REGISTER_WRITE_COMMAND|$0700
VDP_REGISTER_UNUSED_08                                       = VDP_REGISTER_WRITE_COMMAND|$0800
VDP_REGISTER_UNUSED_09                                       = VDP_REGISTER_WRITE_COMMAND|$0900
VDP_REGISTER_HORIZONTAL_INTERRUPT_COUNTER                    = VDP_REGISTER_WRITE_COMMAND|$0A00
VDP_REGISTER_MODE_REGISTER_3                                 = VDP_REGISTER_WRITE_COMMAND|$0B00
VDP_REGISTER_MODE_REGISTER_4                                 = VDP_REGISTER_WRITE_COMMAND|$0C00
VDP_REGISTER_HORIZONTAL_SCROLL_DATA_LOCATION                 = VDP_REGISTER_WRITE_COMMAND|$0D00
VDP_REGISTER_UNUSED_NAMETABLE_PATTERN_GENERATOR_BASE_ADDRESS = VDP_REGISTER_WRITE_COMMAND|$0E00
VDP_REGISTER_AUTO_INCREMENT_VALUE                            = VDP_REGISTER_WRITE_COMMAND|$0F00
VDP_REGISTER_PLANE_SIZE                                      = VDP_REGISTER_WRITE_COMMAND|$1000
VDP_REGISTER_WINDOW_PLANE_HORIZONTAL_EXTENT                  = VDP_REGISTER_WRITE_COMMAND|$1100
VDP_REGISTER_WINDOW_PLANE_VERTICAL_EXTENT                    = VDP_REGISTER_WRITE_COMMAND|$1200
VDP_REGISTER_DMA_LENGTH_LOW_BYTE                             = VDP_REGISTER_WRITE_COMMAND|$1300
VDP_REGISTER_DMA_LENGTH_HIGH_BYTE                            = VDP_REGISTER_WRITE_COMMAND|$1400
VDP_REGISTER_DMA_SOURCE_ADDRESS_L                            = VDP_REGISTER_WRITE_COMMAND|$1500
VDP_REGISTER_DMA_SOURCE_ADDRESS_M                            = VDP_REGISTER_WRITE_COMMAND|$1600
VDP_REGISTER_DMA_SOURCE_ADDRESS_H                            = VDP_REGISTER_WRITE_COMMAND|$1700

VDP_REGISTER_1_EXTRA_VRAM_ENABLE               = %10000000 ; SEGA says always zero.
VDP_REGISTER_1_LEFTMOST_BLANK_ENABLE           = %00100000 ; SEGA says always zero.
VDP_REGISTER_1_HORIZONTAL_INTERRUPT_ENABLE     = %00010000
VDP_REGISTER_1_HV_COUNTER_LATCH_ENABLE         = %00000010
VDP_REGISTER_1_DISPLAY_ENABLE_AND_OVERLAY      = %00000001 ; SEGA says always zero.

VDP_REGISTER_2_TMS9918A_DISPLAY_ENABLE         = %10000000 ; SEGA says always zero.
VDP_REGISTER_2_DISPLAY_ENABLE                  = %01000000
VDP_REGISTER_2_VERTICAL_INTERRUPT_ENABLE       = %00100000
VDP_REGISTER_2_DMA_ENABLE                      = %00010000
VDP_REGISTER_2_PAL_ONLY_240_RESOLUTION_ENABLE  = %00001000
VDP_REGISTER_2_MEGA_DRIVE_VIDEO_MODE_ENABLE    = %00000100 ; SEGA says always one.

VDP_REGISTER_3_TH_LIGHTGUN_INTERRUPT_ENABLE    = %00001000
VDP_REGISTER_3_VERTICAL_SCROLL_2_CELLS         = %00000100
VDP_REGISTER_3_HORIZONTAL_SCROLL_COMPLETELY    = %00000000
VDP_REGISTER_3_HORIZONTAL_SCROLL_8_PIXELS      = %00000010
VDP_REGISTER_3_HORIZONTAL_SCROLL_1_PIXELS      = %00000011

VDP_REGISTER_4_320_PIXEL_WIDE_DISPLAY_ENABLE   = %10000001
VDP_REGISTER_4_VSYNC_IS_NOW_PIXEL_CLOCK_ENABLE = %01000000 ; SEGA says always zero.
VDP_REGISTER_4_HORIZONTAL_SYNC_FREEZE_ENABLE   = %00100000 ; SEGA says always zero.
VDP_REGISTER_4_EXTERNAL_PIXEL_BUS_ENABLE       = %00010000 ; SEGA says always zero.
VDP_REGISTER_4_SHADOW_HIGHLIGHT_MODE_ENABLE    = %00001000
VDP_REGISTER_4_INTERLACE_ENABLE                = %00000010
VDP_REGISTER_4_SONIC_2_DOUBLE_INTERLACE_ENABLE = %00000110

; These are the MSBs of the 16-bit short VDP command words we can send (if you want to access VSRAM you need a longword command!)
                           ;FEDCBA9876543210
                           ;10AAAAAAaaaaaaaa
VDP_WCOMMAND_VRAM_READ   = %0000000000000000
VDP_WCOMMAND_VRAM_WRITE  = %0100000000000000
VDP_WCOMMAND_CRAM_WRITE  = %1100000000000000

; Remember when you use one of these that you should really execute a final VRAM READ on $0000 to clear the topmost mode bits.
; Load a word zero into (MM_VDP_CONTROL+2) that'll do it.
; Don't forget you need to shift your target address up by 16 bits to put it in the high word.
                           ;FEDCBA9876543210FEDBCA9876543210
                           ;10AAAAAAaaaaaaaaXXXXXXXX5432xxAA
VDP_LCOMMAND_VSRAM_READ  = %00000000000000000000000000010000
VDP_LCOMMAND_VSRAM_WRITE = %01000000000000000000000000010000

VDP_LCOMMAND_CRAM_READ   = %00000000000000000000000000100000

; Disable interrupts: set supervisor bit and set mask to 7 to disable all interrupts.
; You MUST disable interrupts before attempting a !TIME write since the vector table will
; cease to exist when the cart bus is isolated.
        macro interrupts_disable
        move    #%0010011100000000,SR                  
        endm
        
; Enable  interrupts: set supervisor bit and set mask to 3 to disable unused interrupts. (Allowing v and hblank.)
        macro interrupts_enable
        move    #%0010001100000000,SR                  
        endm

        macro z80_bus_request_engage
        move.w  #$0100,(MM_Z80_BUS_REQUEST)
        endm

        macro z80_wait_bus_available
\@      btst    #0,(MM_Z80_BUS_REQUEST)
        bne     \@
        endm

        macro z80_bus_request_disengage
        move.w  #$0000,(MM_Z80_BUS_REQUEST)
        endm

        macro z80_reset_engage
        move.w  #$0000,(MM_Z80_RESET)
        endm

        macro z80_reset_disengage
        move.w  #$0100,(MM_Z80_RESET)
        endm

; Using 'A' (data register) as loop counter, copy palette data pointed to by 'B' (address register)
; into palette line 'C' (constant) 
; Assumes VRAM increment of 2.
        macro palette_write
        move.w  #(VDP_WCOMMAND_CRAM_WRITE)|(32*(\3)),(MM_VDP_CONTROL)
        move.w  #15,\1
\@next  move.w  (\2)+,(MM_VDP_DATA)
        dbf     \1,\@next
        endm

        macro YM2612_wait
\@again btst    #YM2612_STATUS_BYTE_BUSY_BIT,(MM_YM2612_ADDRESS_P1)
        bne     \@again
        endm

; Move 'A' (word constant) number of copies of value 'B' (byte constant) into memory pointed at by
; 'C' (address register) ascending using 'D' (data register) as the loop counter.
        macro mem_fill
        move.w  #(\1-1),\4
\@next  move.b  #\2,(\3)+
        dbf     \4,\@next
        endm

; Write the 16-bit value in d0.w to TIME, activating the Game Raccoon cart.
; This isolates the cart from the bus, so any ROM operations will fail at this point!
        macro TIME_write
        movem.l d0/d1,-(a7)
        move.w  d0,(MM_TIME_ADDRESS_BEGIN)
        ; Need to delay for 0.1s
        move.w  #15,d1
\@waitouter        
        move.w  #65535,d0
\@waitinner
        dbf     d0,\@waitinner
        dbf     d1,\@waitouter
        movem.l (a7)+,d0/d1
        endm
        
        macro GAME_RACCOON_isolation_test
        move.l   (MM_STACK_POSITION),d0
        and.l    #1,d0                        ; The cart is not isolated if this can be zero.
        endm