; Game Raccoon Revision 0 Source Code
; Copyright (C) 2019 Mathew Carr

; This program is free software: you can redistribute it and/or modify
; it under the terms of the GNU Affero General Public License Version 3
; with Additional Terms as found in the file
; LICENSE_AGPL_V3_ADDITIONAL_TERMS.txt.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU Affero General Public License for more details.

; You should have received a copy of the GNU Affero General Public License
; along with this program.  If not, see <https://www.gnu.org/licenses/>.
    
;;

        include "megadrive_defines.68k.asm"
        include "included_binaries.68k.asm"

GAME_RACCOON_TIME_WRITE_MAGIC_NUMBER_GR_HANDSHAKE                = $4752        
GAME_RACCOON_TIME_WRITE_MAGIC_NUMBER_REGENERATE_DIRECTORY_CACHE  = $FFFE      
GAME_RACCOON_TIME_WRITE_MAGIC_NUMBER_BOOT_CURRENT_GAME_PARTITION = $FFFF       

; Constants for controller state management:

               ;FEDCBA9876543210
CONTROLLER_U = %0000000000000001
CONTROLLER_D = %0000000000000010
CONTROLLER_L = %0000000000000100
CONTROLLER_R = %0000000000001000
CONTROLLER_B = %0000000000010000
CONTROLLER_C = %0000000000100000
CONTROLLER_A = %0000000001000000
CONTROLLER_S = %0000000010000000
CONTROLLER_M = %0000000100000000
CONTROLLER_Z = %0000001000000000
CONTROLLER_Y = %0000010000000000
CONTROLLER_X = %0000100000000000
CONTROLLER_P = %0100000000000000
CONTROLLER_6 = %1000000000000000

; Active LOW!
; Use these for BTST if we've loaded a controller memory word into a register first.
CONTROLLER_U_BIT = 0
CONTROLLER_D_BIT = 1
CONTROLLER_L_BIT = 2
CONTROLLER_R_BIT = 3
CONTROLLER_B_BIT = 4
CONTROLLER_C_BIT = 5
CONTROLLER_A_BIT = 6
CONTROLLER_S_BIT = 7
CONTROLLER_M_BIT = 8
CONTROLLER_Z_BIT = 9
CONTROLLER_Y_BIT = 10
CONTROLLER_X_BIT = 11
CONTROLLER_P_BIT = 14 ; Down if a controller is present at all.
CONTROLLER_6_BIT = 15 ; Down we if we are using a six button pad. :3

CONTROLLER_U_BIT_btst_l = 0 ; BTST only works on bytes when hitting a memory location for data source.
CONTROLLER_D_BIT_btst_l = 1 ; We need to target the low byte (+1) if we're BTSTing for this!
CONTROLLER_L_BIT_btst_l = 2
CONTROLLER_R_BIT_btst_l = 3
CONTROLLER_B_BIT_btst_l = 4
CONTROLLER_C_BIT_btst_l = 5
CONTROLLER_A_BIT_btst_l = 6
CONTROLLER_S_BIT_btst_l = 7
CONTROLLER_M_BIT_btst_h = 0 ; We need to target the high byte (+0) if we're BTSTing for this!
CONTROLLER_Z_BIT_btst_h = 1
CONTROLLER_Y_BIT_btst_h = 2
CONTROLLER_X_BIT_btst_h = 3
CONTROLLER_P_BIT_btst_h = 6
CONTROLLER_6_BIT_btst_h = 7      
        
        ; The RAM resident code is compiled to run from $FF0000
        ; It'll be copied into position by the entry point.
        ; Only the code part is compiled into the binary object code.
        ; At the address of $FF0000+CODELENGTH, the static RAM memory
        ; map with my variables in begins.
        org     RAM_START
game_raccoon_executable_area_entry_point:        
        jsr     vdp_initialise_registers       
        jsr     vdp_clear_all_cram

        jsr     z80_initialise
        
        jsr     vdp_clear_all_vram               ; Clears the sprite table too, which is held in VRAM of course.
        jsr     vdp_clear_all_vsram              ; Clears the VSRAM.
                
        move.l  #0,(MM_VDP_CONTROL)              ; Clear the extra bits of the VDP control word before we mess with the font!
                
        ; Let's write the font binary into VRAM.
        ; It's 1 bit font so I want to expand every byte into eight pixels.
        ; Every pixel is 4 bits in a Mega Drive tile.
        ; d0 contains the number of longs remaining from the source data.        
        ; Instruct the VDP we want to write some patterns.
        ; Copy my pattern into VRAM at $36C0. ($3680 is the first tile, but it's a SPACE character! Also bear in mind we want double-sized chars.)
        move.w  #(VDP_WCOMMAND_VRAM_WRITE|$36C0),(MM_VDP_CONTROL)
        lea     binary_triodx_font_1bit_bin,a0                                             ; Read the 1bpp data from this address.
        lea     font_1bpp_four_pixel_groups_mapping_to_4bpp_words,a1                       ; Retrieve the expanded pixel data from this address.
        move.w  #(binary_triodx_font_1bit_bin_length_words-1),d0                           ; Loop for how many words are in the 1bpp font.
font_1bpp_load_next_word:        
        move.w  (a0)+,d2                                                                   ; Load the next word of 1bpp graphics into d2.
        move.w  #(4-1),d1                                                                  ; There are four four-bit sequences in a word of 1bpp data.
font_1bpp_load_next_quadbits:        
        move.w  d2,d3                                                                      ; Copy the loaded word into d3 for ANDing.
        and.w   #%00001111,d3                                                              ; Get the least significant four bits.
        lsl.w   #1,d3                                                                      ; We're indexing into a table of words!!
        move.w  (a1,d3.w),(MM_VDP_DATA)                                                    ; Write the computed word to the VDP pattern RAM.
        ror.w   #4,d2                                                                      ; Move the next least significant four bits into the lowest.
        dbf     d1,font_1bpp_load_next_quadbits
        dbf     d0,font_1bpp_load_next_word
     
        jsr     screen_clear_plane_a_to_space
        jsr     screen_clear_plane_b_to_space
        
        lea     (palette_0),a0
        palette_write d0,a0,0
        lea     (palette_1),a0
        palette_write d0,a0,1
        lea     (palette_2),a0
        palette_write d0,a0,2
        lea     (palette_3),a0
        palette_write d0,a0,3
        
        jsr     screen_clear_plane_a_to_space
        move.l  #(2<<13),d2
        move.l  #1,d0
        move.l  #8,d1
        lea     (string_welcome1),a0
        jsr     write_string
                
        move.w  #GAME_RACCOON_TIME_WRITE_MAGIC_NUMBER_GR_HANDSHAKE,d0
        TIME_write
handshake_loop_until_deisolated:
        GAME_RACCOON_isolation_test
        bne     handshake_loop_until_deisolated
        
        jsr     screen_clear_plane_a_to_space
        move.l  #(2<<13),d2
        move.l  #1,d0
        move.l  #8,d1
        lea     (string_sdcard),a0
        jsr     write_string
        
        move.w  #GAME_RACCOON_TIME_WRITE_MAGIC_NUMBER_REGENERATE_DIRECTORY_CACHE,d0
        TIME_write
initialise_loop_until_deisolated:
        GAME_RACCOON_isolation_test
        bne     initialise_loop_until_deisolated
        
        move.l  #0,menu_state_cursor_location
        move.l  #0,menu_state_scroll_top_row_shown
        
        move.l  #0,d0
        jsr     directory_cache_assemble_options_list                     
        jsr     screen_draw
        
        jsr     controllers_reset
        jsr     controllers_read
        jsr     controllers_derive
        
loop:
        jsr     vcounter_wait
        
        jsr     controllers_read
        jsr     controllers_derive
        
        btst    #CONTROLLER_S_BIT_btst_l,pad_1_pressed+1
        beq     GAME_RACCOON_game_partition_boot_sequence

        btst    #CONTROLLER_C_BIT_btst_l,pad_1_pressed+1
        beq     loop_change_to_parent_directory
        
        btst    #CONTROLLER_B_BIT_btst_l,pad_1_pressed+1
        bne     loop_b_idle        
        ; Check we're on a directory or a game.
        move.l  menu_state_cursor_location,d0
        sub.l   directory_cache_items_list_status_struct_no_directories_inc_parent_row,d0  ; d0 is >= 0 if we're on a file.
        blt     loop_b_selected_directory     
        ; You've selected a file to install.        
        jsr     screen_clear_plane_a_to_space
        move.l  #(0<<13),d2
        move.l  #10,d0
        move.l  #8,d1
        lea     (string_men1),a0
        jsr     write_string
        move.l  #8,d0
        move.l  #10,d1
        lea     (string_men2),a0
        jsr     write_string
        ; We want to install the file that has index ((directory_cache_items_list_begin + 4*cursor_position)&$1FFFFF) >> 8
        lea     (directory_cache_items_list_begin),a0
        move.l  menu_state_cursor_location,d0                              ; d0 = cursor
        lsl.l   #2,d0                                                      ; d0 = cursor*4
        move.l  (a0,d0),d0                                                 ; Full entry from the in-ram sorted list.
        and.l   #$1FFFFF,d0                                                ; Strip the rest of the pointer away.
        lsr.l   #8,d0                                                      ; Index number recovered
        jsr     GAME_RACCOON_install_game_at_directory_cache_record_index_d0     
        jmp     loop_display_dirty_redraw
loop_b_selected_directory:
        ; If we're not on the root item, and
        ; If the special parent is there and you picked item zero, then go parent.
        move.l  directory_cache_items_list_status_struct_current_index,d0
        beq     loop_b_no_special_parent_handling
        move.l  menu_state_cursor_location,d0  
        bne     loop_b_no_special_parent_handling
loop_change_to_parent_directory:
        move.l  directory_cache_items_list_status_struct_parent_index,d0  
        jsr     directory_cache_assemble_options_list
        jmp     loop_display_dirty_redraw
        ; You selected a directory:
loop_b_no_special_parent_handling:
        lea     (directory_cache_items_list_begin),a0
        move.l  menu_state_cursor_location,d0                              ; d0 = cursor
        lsl.l   #2,d0                                                      ; d0 = cursor*4
        move.l  (a0,d0),d0                                                 ; Full entry from the in-ram sorted list.
        and.l   #$1FFFFF,d0                                                ; Strip the rest of the pointer away.
        lsr.l   #8,d0                                                      ; Index number recovered
        jsr     directory_cache_assemble_options_list
        jmp     loop_display_dirty_redraw
loop_b_idle:        
        
        
        ; Control cursor.        
        move.l  menu_state_cursor_location,d6                              ; d6 is unmodified cursor location.
        move.l  d6,d5                                                      ; d5 is cursor location possibly modified.
        btst    #CONTROLLER_U_BIT_btst_l,pad_1_pressed+1
        bne     loop_up_idle
        sub.l   #1,d5
loop_up_idle:
        btst    #CONTROLLER_D_BIT_btst_l,pad_1_pressed+1
        bne     loop_down_idle
        add.l   #1,d5
loop_down_idle:
        cmp.l   #0,d5
        blt     loop_cursor_action_bad
        cmp.l   directory_cache_items_list_status_struct_current_length,d5
        bge     loop_cursor_action_bad
        jmp     loop_cursor_action_good
loop_cursor_action_bad:
        move.l  d6,d5
loop_cursor_action_good:
        cmp.l   d5,d6
        beq     loop_cursor_no_action                
        move.l  d5,menu_state_cursor_location                              
        ; Handle scrolling.
        ; if cursor <= scroll_toprow: scroll_toprow = cursor-1
        ; if cursor >= (scroll_toprow + screensize-1): scroll_toprow = cursor - (screensize-2)
        ; if scroll_toprow > length - screensize: scroll_toprow = length - screensize
        ; if scroll_toprow < 0: scroll_toprow = 0
SCREEN_SIZE_ROWS = 8        
        move.l  menu_state_cursor_location,d0
        move.l  menu_state_scroll_top_row_shown,d1  ; d1 contains the modified top row shown as the algorithm progresses.
        ; Rule 1
        cmp.l   d1,d0
        bgt     cursor_not_off_top_of_list
        move.l  d0,d1
        sub.l   #1,d1
cursor_not_off_top_of_list:        
        ; Rule 2
        move.l  d1,d2
        add.l   #(SCREEN_SIZE_ROWS-1),d2
        cmp.l   d2,d0
        blt     cursor_not_off_bottom_of_list
        move.l  d0,d1
        sub.l   #(SCREEN_SIZE_ROWS-2),d1
cursor_not_off_bottom_of_list:        
        ; Rule 3
        move.l  directory_cache_items_list_status_struct_current_length,d2
        sub.l   #SCREEN_SIZE_ROWS,d2
        cmp.l   d2,d1
        ble     scroll_not_off_bottom_of_list
        move.l  d2,d1
scroll_not_off_bottom_of_list:
        ; Rule 4
        cmp.l   #0,d1
        bge     scroll_not_off_top_of_list
        move.l  #0,d1
scroll_not_off_top_of_list:
        move.l  d1,menu_state_scroll_top_row_shown

loop_display_dirty_redraw:
        jsr     screen_clear_plane_a_to_space
        jsr     screen_draw
        
loop_cursor_no_action:

        
        
        jmp     loop
        
        align   2        
       
; Draw the menu list into plane a using the cursor state
; and current directory record cache list.
screen_draw:
        ; Draw a max of SCREEN_SIZE_ROWS.
        ; Draw rows from row menu_state_scroll_top_row_shown, but not equal or greater than directory_cache_items_list_status_struct_current_length.
        ; No rows to draw in d7.
        move.l  menu_state_scroll_top_row_shown,d5
        
        move.l  directory_cache_items_list_status_struct_current_length,d7
        sub.l   d5,d7                                                                     ; d7 = number of rows available to draw
        cmp.l   #SCREEN_SIZE_ROWS,d7
        ble     screen_draw_number_of_rows_acceptable
        move.l  #SCREEN_SIZE_ROWS,d7                                                      ; Reduce number of rows to screen size.
screen_draw_number_of_rows_acceptable:
        sub.l   #1,d7                                                                     ; Remove one for dbf.

        lea     (directory_cache_items_list_begin),a1
        move.l  d5,d0
        lsl.l   #2,d0
        lea     (a1,d0),a1                                                                ; Advance a1 by 4* the cursor location.
        
        move.l  #1,d0 
        move.l  #0,d1 
        move.l  #(1<<13),d2                                                               ; Text colour is blue to start.
        
        move.l  directory_cache_items_list_status_struct_no_directories_inc_parent_row,d6 ; d6 is total number of directories including parent row.
        sub.l   d5,d6                                                                     ; d6 is the number of exposed directories on current scroll.  
        bgt     screen_draw_calculate_directory_total_nonnegative_shown
        move.l  #0,d6                                                                     ; No rows in directory colour.
        move.l  #(0<<13),d2                                                               ; Change to white for files.
screen_draw_calculate_directory_total_nonnegative_shown:        
        
screen_draw_draw_next_row:
        move.l  (a1)+,a0
        jsr     write_string_from_final_slash
        add.w   #2,d1
        sub.l   #1,d6
        bne     screen_draw_draw_next_row_dont_switch_to_white
        move.l  #(0<<13),d2                                                               ; Change to white for files when d6 == 0
screen_draw_draw_next_row_dont_switch_to_white:        
        dbf     d7,screen_draw_draw_next_row
        
        ; Drawing the chevron.
        move.l  #0,d0 
        move.l  menu_state_cursor_location,d1
        sub.l   d5,d1
        lsl.l   #1,d1
        lea     (string_chevron),a0
        move.l  #(2<<13),d2
        jsr     write_string_from_final_slash
        
        rts
        
string_chevron:
        dc.b   ">",0        
string_men1:
        dc.b   "Please wait while",0
string_men2:
        dc.b   "MEN install this game",0
string_welcome1:
        dc.b   "GAME RACCOON! Handshaking...",0
string_sdcard:
        dc.b   "Reading SD Card!",0
        
        align  2
                
screen_clear_plane_a_to_space:
        ; Fill the name table entries in plane A with the space character from the loaded font.
        lea     (MM_VDP_DATA),a6        
        move.w  #$01B4,d1        
        move.w  #((64*32)-1),d0
        move.w  #(VDP_WCOMMAND_VRAM_WRITE|$0000),(MM_VDP_CONTROL)         
screen_clear_plane_a_to_space_loop:
        move.w  d1,(a6)
        dbf     d0,screen_clear_plane_a_to_space_loop
        rts        
                
screen_clear_plane_b_to_space:
        ; Fill the name table entries in plane B with the space character from the loaded font.
        lea     (MM_VDP_DATA),a6        
        move.w  #$01B4,d1        
        move.w  #((64*32)-1),d0
        move.w  #(VDP_WCOMMAND_VRAM_WRITE|$2000),(MM_VDP_CONTROL)         
screen_clear_plane_b_to_space_loop:
        move.w  d1,(a6)
        dbf     d0,screen_clear_plane_b_to_space_loop
        rts                
                
vcounter_wait:
        ; Wait for the VDP to be less than N, then MORE than N.
        move.b  #200,d0
vw1:    cmp.b  (MM_VDP_HV_COUNTER),d0
        bge    vw1
vw2:    cmp.b  (MM_VDP_HV_COUNTER),d0
        blt    vw2
        rts
        
                
; With the index of a directory cache record in d0.l, (top 16-bits must be zero!!)
; populate directory_cache_items_list_begin with longword-sized addresses of all the items
; This is an area of memory with enough space to store
; 1024 longword items. If you have more than that in one folder,
; you're gonna be out of luck. Don't do that!!
; that ought to appear in its menu.
; string_parent is the '..' parent option, shown first only
; if the d0 is not the root entry.
; The list will have the format:
; PARENT ELEMENT MARKER (if not root)
; DIRa
; DIRb
; DIRc sorted alphabetically
; DIRd
; FILEa
; FILEb 
; FILEc
; FILEd sorted alphabetically
; FILEe
; 
; Sets the following:
; directory_cache_items_list_status_struct_current_index 
; directory_cache_items_list_status_struct_current_length is set to the length of the list = (no_dirs + no_files + 1 if nonparent)
; directory_cache_items_list_status_struct_no_directories        
; directory_cache_items_list_status_struct_no_files              
; directory_cache_items_list_status_struct_parent_index          
directory_cache_assemble_options_list:
        movem.l directory_cache_assemble_options_list_regs,-(a7)
        
        and.l   #$0000FFFF,d0
        move.l  d0,directory_cache_items_list_status_struct_current_index
        
        move.l  #0,d3                                                             ; d3.w = number of items located. 
        
        move.l  d3,menu_state_cursor_location
        move.l  d3,menu_state_scroll_top_row_shown
        move.l  d3,directory_cache_items_list_status_struct_no_directories        ; Clear some fields with d3, the upper half needs initialising.
        move.l  d3,directory_cache_items_list_status_struct_no_files
        move.l  d3,directory_cache_items_list_status_struct_current_length
        move.l  d3,d1
        
        lea     directory_cache_items_list_begin,a1                               ; Write to the global list.
        
        ; If this is not the root element, add a string_parent entry to the beginning of the list.
        cmp.l   #0,d0
        beq     directory_cache_assemble_options_list_no_parent
        move.l  #string_parent,(a1)+
        add.w   #1,d3                                                             ; d3++, stored an item.
directory_cache_assemble_options_list_no_parent:


        jsr     directory_cache_convert_index_to_address                          ; a6 = address of the current directory record.
        move.l  a0,a6                                                             ; ^
        
        move.w  (DIRECTORY_CACHE_OFFSET_TO_PARENT_INDEX,a6),directory_cache_items_list_status_struct_parent_index+2 ; Storing into longword.
        
        ; Copy the addresses of the directory entries that belong to this item into the list a1 onwards, advancing a1.
        move.w  (DIRECTORY_CACHE_OFFSET_TO_CONTENTS_BEGIN,a6),d0                  ; d0.l = index of directory cache record with contents, starting with directories
        jsr     directory_cache_convert_index_to_address                          ; a0 = base address of first directory/file in this directory, all items are 256+ stride from this
        ; DIRECTORIES LOOP.
        move.l  a1,a3                                                             ; a3 stores address of output list (possibly had parent added) before directories are added.        
        move.w  (DIRECTORY_CACHE_OFFSET_TO_DIRECTORY_COUNT,a6),d1                 ; d2.w = d1.w = number of directories this directory entry possesses.
        move.l  d1,directory_cache_items_list_status_struct_no_directories      
        add.w   d1,d3                                                             ; d3 += no directories.
        move.l  d3,directory_cache_items_list_status_struct_no_directories_inc_parent_row
        move.w  d1,d2                                                             ; Need it in d2.w to control the length of the sort below.
        cmp.w   #0,d1                                                             ; Zero children: skip the loop. (Can't use dbf on a zero list.)
directory_cache_assemble_options_list_next_directory:     
        beq     directory_cache_assemble_options_list_directories_added
        move.l  a0,(a1)+                                                          ; Move the current address found into the list.
        lea     (256,a0),a0                                                       ; Move a0 to the next directory cache record, the next directory that belongs to this list.
        sub.w   #1,d1                                                             ; Decrement remaining number of directories to identify.
        jmp     directory_cache_assemble_options_list_next_directory
directory_cache_assemble_options_list_directories_added:
        ; a3 points to the slot that was vacant before we added any directories
        ; a1 points to the next vacant slot in the collection list
        ; a1 = a3 + 4*(no directories)
        
        ; Now we need to alphabetically sort a list of addresses of d2.w directories starting at a3.
        jsr     directory_cache_sort_address_list_alphabetically
        ; Alright! The addresses in the directory part of the output list are sorted alphabetically!
        
        ; FILES LOOP.
        move.l  a1,a3                                                             ; a3 stores address of output list (possibly had parent added) before files are added.        
        move.w  (DIRECTORY_CACHE_OFFSET_TO_FILE_COUNT,a6),d1                      ; d2.w = d1.w = number of files this directory entry possesses.
        move.l  d1,directory_cache_items_list_status_struct_no_files    
        add.w   d1,d3                                                             ; d3 += no files.
        move.w  d1,d2                                                             ; Need it in d2.w to control the length of the sort below.
        cmp.w   #0,d1                                                             ; Zero children: skip the loop. (Can't use dbf on a zero list.)
directory_cache_assemble_options_list_next_file:     
        beq     directory_cache_assemble_options_list_files_added
        move.l  a0,(a1)+                                                          ; Move the current address found into the list.
        lea     (256,a0),a0                                                       ; Move a0 to the next directory cache record, the next file that belongs to this list.
        sub.w   #1,d1                                                             ; Decrement remaining number of files to identify.
        jmp     directory_cache_assemble_options_list_next_file
directory_cache_assemble_options_list_files_added:
        ; a3 points to the slot that was vacant before we added any files
        ; a1 points to the next vacant slot in the collection list
        ; a1 = a3 + 4*(no files)
        
        ; Now we need to alphabetically sort a list of addresses of d2.w directories starting at a3.
        jsr     directory_cache_sort_address_list_alphabetically
        ; Alright! The addresses in the file part of the output list are sorted alphabetically!
                        
        move.l  d3,directory_cache_items_list_status_struct_current_length        ; Store the total number of entries in the list in the global variable.
        
        movem.l (a7)+,directory_cache_assemble_options_list_regs
        rts        
directory_cache_assemble_options_list_regs reg d0-d7/a0-a6

; Given the index of a directory entry in d0.l,
; return the address of the entry in a0.
directory_cache_convert_index_to_address:
        movem.l directory_cache_convert_index_to_address_regs,-(a7)
        
        lsl.l   #8,d0                                                     ; Multiply the index by 256 (size of directory entry) to get offset into directory cache.
        lea     (DIRECTORY_CACHE_BASE),a0
        lea     (a0,d0.l),a0                                              ; a0 = BASE + 256*index
        
        movem.l (a7)+,directory_cache_convert_index_to_address_regs
        rts        
directory_cache_convert_index_to_address_regs reg d0

; From a3 for d2.w number of longwords, sort the directory cache records by their filenames.
; No return - sort the list in place. IF YOU DARE.
directory_cache_sort_address_list_alphabetically:
        movem.l directory_cache_sort_address_list_alphabetically_regs,-(a7)
        
directory_cache_sort_address_list_alphabetically_next_iteration:        
        cmp.w   #2,d2                                                                            ; If there is one or zero items, then return.
        bge     directory_cache_sort_address_list_alphabetically_continue_iteration
        movem.l (a7)+,directory_cache_sort_address_list_alphabetically_regs
        rts       
directory_cache_sort_address_list_alphabetically_continue_iteration:

        ; Alright, we're going to find the 'least' item as we go through d2 items.
        ; After finding the 'least' item, swap its value with the one in a3 and
        ; decrement d2 and increment a3, since we have placed the least item in
        ; the appropriate location.
        ; Then repeat to sort the list until there are no more items to sort.
        ; a2 is a sliding cursor going from a3+4 onwards for d2-2 advances.
        move.w  d2,d3
        sub.w   #2,d2
        move.l  a3,a2
directory_cache_sort_address_list_alphabetically_advance_sliding_cursor:        
        lea     (4,a2),a2       
        move.l  (a2),a1
        move.l  (a3),a0
        jsr     strcmp                                                                           ; CC equals a0 <?> a1.
        bge     directory_cache_sort_address_list_alphabetically_no_swap
        move.l  a0,(a2)                                                                          ; Swap the contents of (a3) and (a2)
        move.l  a1,(a3)                                                                          ; Swap ^        
directory_cache_sort_address_list_alphabetically_no_swap:
        dbf     d2,directory_cache_sort_address_list_alphabetically_advance_sliding_cursor
        
        move.w  d3,d2                                                                            ; The length of the unsorted list is one shorter.
        sub.w   #1,d2  
        lea     (4,a3),a3
        jmp     directory_cache_sort_address_list_alphabetically_next_iteration                  ; Sort this list!
directory_cache_sort_address_list_alphabetically_regs reg d2-d3/a0-a3

; Set the condition code to indicate whether the string pointed to by a0 is
; greater or less than the one in a1.

; Compare the characters in both and advance.
; If they're different, then return and the result is in the condition code.
; If they're the same, check for zero. If they're the same and zero, they're equal, return that in the condition code.
; Otherwise they're the same but nonzero, so repeat the test.
strcmp:
        movem.l strcmp_return_regs,-(a7)
strcmp_next_char:
        cmp.b   (a0)+,(a1)+
        bne     strcmp_return        
        cmp.b   #0,-1(a0)
        beq     strcmp_return        
        jmp     strcmp_next_char        
strcmp_return:
        movem.l (a7)+,strcmp_return_regs
        rts        
strcmp_return_regs reg a0-a1

; a0 dst
; a1 src   
strcpy: 
        movem.l strcpy_regs,-(a7)
strcpy_repeat:
        move.b  (a1)+,(a0)+
        bne     strcpy_repeat
        
        movem.l (a7)+,strcpy_regs
        rts        
strcpy_regs reg a0-a1


DIRECTORY_CACHE_BASE                      = $200000
DIRECTORY_CACHE_ENTRY_SIZE_BYTES          =    $100       
DIRECTORY_CACHE_ENTRY_SIZE_WORDS          = (DIRECTORY_CACHE_ENTRY_SIZE_BYTES/2)
DIRECTORY_CACHE_OFFSET_TO_DIRECTORY_COUNT = (DIRECTORY_CACHE_ENTRY_SIZE_BYTES-(4*2))
DIRECTORY_CACHE_OFFSET_TO_FILE_COUNT      = (DIRECTORY_CACHE_ENTRY_SIZE_BYTES-(3*2))
DIRECTORY_CACHE_OFFSET_TO_CONTENTS_BEGIN  = (DIRECTORY_CACHE_ENTRY_SIZE_BYTES-(2*2))
DIRECTORY_CACHE_OFFSET_TO_PARENT_INDEX    = (DIRECTORY_CACHE_ENTRY_SIZE_BYTES-(1*2))
                
        align   2

; Arguments, same as below.
; Only draws from the first character after the final slash!
; There has to be at least one slash!!
write_string_from_final_slash:
        move.l  a1,-(a7)
        move.l  a0,a1

write_string_from_final_slash_looking_for_end:        
        cmp.b   #0,(a0)+
        bne     write_string_from_final_slash_looking_for_end
             
write_string_from_final_slash_rewind:
        cmp.l   a0,a1
        beq     write_string_from_final_slash_no_slash
        cmp.b   #'/',-(a0)
        bne     write_string_from_final_slash_rewind
        cmp.b   #' ',(a0)+                                             ; Dummy operation to advance the cursor back after the slash again.
        
write_string_from_final_slash_no_slash:
        move.l  (a7)+,a1
        
; Arguments:
;  d0.w = x
;  d1.w = y of top half
;  d2.w = palette (must be << 13)
;  a0   = pointer to null terminated string        
write_string:
        movem.l write_string_regs,-(a7)
        
        lea     (MM_VDP_DATA),a6

        ; Since the VRAM counter autoincrements, it's best to draw the string in two strips!
        move.l  a0,a1                                                  ; Keep original pointer.
        
        add.w   #$01B4-(' '*2),d2                                      ; d2 contains base tile number and palette value.
        
        ; Construct the address in VRAM of the destination.
        lsl.w   #6,d1                                                  ; Multiply the y coordinate by the screen width in tiles (512/8).
        or.w    d1,d0                                                  ; d0 is tile index in map.
        lsl.w   #1,d0                                                  ; Multiply by two because entries are words. Since we're writing into a table at $0000 we're done.
        or.w    #(VDP_WCOMMAND_VRAM_WRITE|$0000),d0                    ; Write into PLANE A.
        move.w  d0,(MM_VDP_CONTROL)                                    ; Target tile X Y in PLANE A.
        
write_string_top_next_character:
        move.w  #0,d3
        move.b  (a0)+,d3                                               ; Character value.
        beq     write_string_top_complete      
        lsl.b   #1,d3                                                  ; Tile value respecting two-halves. ; $01B4 is the space tile index, as we've loaded the text into the lowest vacant space.
        add.w   d2,d3                                                  ; d3 is tile index - 0 priority, 0 palette, 0 vh flip.
        move.w  d3,(a6)                                                ; Write name table entry to VDP data register.
        jmp     write_string_top_next_character
write_string_top_complete:        

        add.w   #128,d0
        move.w  d0,(MM_VDP_CONTROL)                                    ; Target tile X Y+1 in PLANE A.
        
write_string_bottom_next_character:
        move.w  #0,d3
        move.b  (a1)+,d3                                               ; Character value.
        beq     write_string_bottom_complete      
        lsl.b   #1,d3                                                  ; Tile value respecting two-halves. ; $01B4 is the space tile index, as we've loaded the text into the lowest vacant space.
        add.w   d2,d3                                                  ; d3 is tile index - 0 priority, 0 palette, 0 vh flip.
        add.w   #1,d3                                                  ; Add one for bottom half.
        move.w  d3,(a6)                                                ; Write name table entry to VDP data register.
        jmp     write_string_bottom_next_character
write_string_bottom_complete:        
        
write_string_complete:        
        movem.l (a7)+,write_string_regs
        rts
write_string_regs reg d0-d7/a0-a6
        
; DONT PUT CONSTANTS IN THE GAME RACCOON'S RAM RESIDENT AREA!!!        
; LIKE THIS:::: YOU CAN PUT THEM IN THE BINARIES INSTEAD. 
; THEN THEYLL BE ROM RESIDENT!!
  
; Read four bits from the source data and you'll get a VDP word to write.
font_1bpp_four_pixel_groups_mapping_to_4bpp_words:
        dc.w    %0000000000000000
        dc.w    %0001000000000000
        dc.w    %0000000100000000
        dc.w    %0001000100000000
        dc.w    %0000000000010000
        dc.w    %0001000000010000
        dc.w    %0000000100010000
        dc.w    %0001000100010000
        dc.w    %0000000000000001
        dc.w    %0001000000000001
        dc.w    %0000000100000001
        dc.w    %0001000100000001
        dc.w    %0000000000010001
        dc.w    %0001000000010001
        dc.w    %0000000100010001
        dc.w    %0001000100010001
        
palette_0:
        dc.w    $0000,$0eee,$0282,$04E4,$0e00,$0e60,$0ea2,$0eee,$0000,$0224,$0228,$044E,$0000,$0244,$0288,$04EE
palette_1:
        dc.w    $0e00,$0ee0,$0ea2,$0eee,$0444,$0464,$0666,$0868,$0888,$08A8,$0AAA,$0CAC,$0CCC,$0CEC,$0EEE,$0EEE
palette_2:
        dc.w    $0000,$00ee,$0222,$0422,$0422,$0422,$0644,$0846,$0866,$0866,$0A8A,$0C8A,$0CAA,$0CAA,$0EAA,$0EEE
palette_3:
        dc.w    $0e00,$0e0e,$0ea2,$0eee,$0e22,$0e22,$0e44,$0e46,$0e66,$0e66,$0e8A,$0e8A,$0eAA,$0eAA,$0EAA,$0EEE
        
        
        
        align   2        
; Put the VDP into a good state. VBlank interrupts enabled. HBlank disabled. 320x224 resolution.
vdp_initialise_registers:
        lea     (MM_VDP_CONTROL),a6
        move.w  #(VDP_REGISTER_MODE_REGISTER_1|$00),(a6)
        move.w  #(VDP_REGISTER_MODE_REGISTER_2|VDP_REGISTER_2_VERTICAL_INTERRUPT_ENABLE|VDP_REGISTER_2_DISPLAY_ENABLE|VDP_REGISTER_2_MEGA_DRIVE_VIDEO_MODE_ENABLE),(a6)
        move.w  #(VDP_REGISTER_MODE_REGISTER_3|$00),(a6)
        move.w  #(VDP_REGISTER_MODE_REGISTER_4|VDP_REGISTER_4_320_PIXEL_WIDE_DISPLAY_ENABLE),(a6)

        move.w  #(VDP_REGISTER_AUTO_INCREMENT_VALUE|$02),(a6) ; We're going to assume this all over the place. If you change it away, CHANGE IT BACK.

        move.w  #(VDP_REGISTER_PLANE_A_NAME_TABLE_LOCATION|%00000000),(a6)            ; NT A = $0000-$1000
        move.w  #(VDP_REGISTER_WINDOW_NAME_TABLE_LOCATION|%00000100),(a6)             ; W B  = $1000-$2000
        move.w  #(VDP_REGISTER_PLANE_B_NAME_TABLE_LOCATION|%00000001),(a6)            ; NT B = $2000-$3000
        move.w  #(VDP_REGISTER_HORIZONTAL_SCROLL_DATA_LOCATION|%00001100),(a6)        ; HS   = $3000-$3400
        move.w  #(VDP_REGISTER_SPRITE_TABLE_LOCATION|%00011010),(a6)                  ; ST   = $3400-$3680

        move.w  #(VDP_REGISTER_BACKGROUND_COLOUR|$00),(a6)
        move.w  #(VDP_REGISTER_PLANE_SIZE|$01),(a6)                                   ; A and B are both 64x 32y.

        move.w  #(VDP_REGISTER_WINDOW_PLANE_HORIZONTAL_EXTENT|$00),(a6)               ; Full zero prevents window plane display.
        move.w  #(VDP_REGISTER_WINDOW_PLANE_VERTICAL_EXTENT|$00),(a6)

        move.w  #(VDP_REGISTER_UNUSED_SPRITE_PATTERN_GENERATOR_BASE_ADDRESS|$00),(a6)
        move.w  #(VDP_REGISTER_UNUSED_08|$00),(a6)
        move.w  #(VDP_REGISTER_UNUSED_09|$00),(a6)
        move.w  #(VDP_REGISTER_UNUSED_NAMETABLE_PATTERN_GENERATOR_BASE_ADDRESS|$00),(a6)
        rts

; Clear the entire contents of the ordinary VRAM. (As opposed to the CRAM or VSRAM.)
vdp_clear_all_vram:
        ; Write the longword zero into MM_VDP_CONTROL to reset the high bits of VDP write address.
        move.l  #0,(MM_VDP_CONTROL)
        ; Set up writing VRAM from VRAM start, incrementing by words.
        move.w  #(VDP_WCOMMAND_VRAM_WRITE),(MM_VDP_CONTROL)
        move.w  #(VRAM_SIZE_WORDS-1),d0
        move.w  #0,d1
vdp_clear_all_vram_next_word:
        move.w  d1,(MM_VDP_DATA)
        dbf     d0,vdp_clear_all_vram_next_word
        rts

; Clear the entire contents of the VSRAM.
vdp_clear_all_vsram:
        ; Set up writing VSRAM from VSRAM start, incrementing by words.
        move.l  #(VDP_LCOMMAND_VSRAM_WRITE),(MM_VDP_CONTROL)
        move.w  #(VSRAM_SIZE_WORDS-1),d0
        move.w  #0,d1
vdp_clear_all_vsram_next_word:
        move.w  d1,(MM_VDP_DATA)
        dbf     d0,vdp_clear_all_vsram_next_word
        move.l  #0,(MM_VDP_CONTROL) ; Need to clear the second word of VDP_CONTROL for expected behaviour.
        rts

; Clear the CRAM to black for all four palettes.
vdp_clear_all_cram:
        ; Write the longword zero into MM_VDP_CONTROL to reset the high bits of VDP write address.
        move.l  #0,(MM_VDP_CONTROL)
        ; Set up writing CRAM from CRAM start, incrementing by words.
        move.w  #(VDP_WCOMMAND_CRAM_WRITE),(MM_VDP_CONTROL)
        move.w  #(CRAM_SIZE_WORDS-1),d0
        move.w  #0,d1
vdp_clear_all_cram_next_word:
        move.w  d1,(MM_VDP_DATA)
        dbf     d0,vdp_clear_all_cram_next_word
        rts

; Make the z80 inert for the time being.
; According to the sega documentation, the correct sequence for loading the Z80 program is:
;    (1) BUS REQ ON
;    (2) BUS RESET OFF
;    (3) 68K copies program into Z-80 S-RAM
;    (4) BUS RESET ON
;    (5) BUS REQ OFF
;    (6) BUS RESET OFF
z80_initialise:
        z80_bus_request_engage
        z80_reset_disengage
        z80_wait_bus_available

        lea     MM_Z80_MEMORY_SPACE,a0

        move.b  #$C3,(a0)+                         ; Put a [jp $0000] at $0000.
        move.b  #$00,(a0)+
        move.b  #$00,(a0)

        z80_reset_engage
        z80_bus_request_disengage
        z80_reset_disengage
        rts

; We're going to get a fine collection of system library routines.
; Plonk them together and then maybe we can refactor them into a separate file later.
; (It's 16th January 2014 today, in case future Matt comes back and sees that he never
; bothered to refactor the controller reading code WHUPS.)
; It's now 25th December 2019, and I hadn't touched Mega Drive code between then and
; now! I'm adapting the Instrument Maker code for the Game Raccoon boot ROM, and
; one factor that's giving me pause about refactoring this code is that it relies on
; RAM globals, and the GR is going to have code in RAM, and requires copying code to
; RAM to run it during bus isolation, so it's simplest to leave the refactoring until
; later.
; Making it 1 idle, 0 pressed for the current states.

; Reset the controllers to known unpressed state.
controllers_reset:
        move.l  #0,pad_1_curr                                  ; This should clear both low and high words! (Curr and prev are stored consecutively.)
        move.l  #0,pad_2_curr
        jsr     controllers_derive
        rts

controllers_read:
        lea     MM_CONTROLLER_1_DATA+1,a0                      ; a0 is the memory mapped register for the data source.
        lea     pad_1_curr,a1                                  ; a1 is the pad status struct in RAM.
        
        move.w  #1,d3                                          ; Go through two pads.

controllers_read_next_controller:
        move.w  (a1),2(a1)                                     ; Copy current state to previous state.

        ; We want to read both controller 1 and controller 2. We do this in a loop because the mm registers are close.
        move.b  #$40,MM_CONTROLLER_DATA_TO_CONTROL_OFFSET(a0)  ; Set direction of bits of port 1.

        ; We're going to try and read the controller port in such a way that's compatible with both the 3 button and 6 button pad.
        ; https://github.com/jonthysell/SegaController/wiki/How-To-Read-Sega-Controllers
                                                               ; Enter this with TH low. (Cycle state 0.)
        
        move.b  #$40,(a0)                                      ; Make TH high. (Cycle state 1.)
        nop                                                    ; Allow the controller to respond.
        nop

        move.b  #$00,(a0)                                      ; Make TH low.  (Cycle state 2.)
        nop                                                    ; Allow the controller to respond.
        nop
        move.b  (a0),d1                                        ; d1 = xxxxxxxxxxSA!!xx --- exclamation marks are double LOW if a controller is present.
        move.b  d1,d2                                          ; d2 = 0000000000001100
        and.b   #%00001100,d2                                  ; d2 = zero if controller present, else no controller.
        beq     controllers_read_present_continue
        move.w  #%1111111111111111,d0                          ; d0 = 1111111111111111 controller fully idle, not present
        jmp     controllers_read_done                          ; return d0
controllers_read_present_continue:
        move.w  #%1011111111111111,d0                          ; d0 = 1011111111111111 set the controller present bit
        lsl.b   #2,d1                                          ; d1 =         SAxxxxxx
        or.b    #%00111111,d1                                  ; d1 =         SA111111 --- active LOW.
        move.b  d1,d0                                          ; d0 = 10111111SA111111

        move.b  #$40,(a0)                                      ; Make TH high. (Cycle state 3.)
        nop                                                    ; Allow the controller to respond.
        nop        
        move.b  (a0),d1                                        ; d1 =         xxCBRLDU
        or.b    #%11000000,d1                                  ; d1 =         11CBRLDU                
        and.b   d1,d0                                          ; d0 = 10111111SACBRLDU --- active LOW.

        move.b  #$00,(a0)                                      ; Make TH low.  (Cycle state 4.)
        nop                                                    ; Allow the controller to respond.
        nop
        move.b  (a0),d1                                        ; d1 =         xxxxxx!! --- exclamation marks are double LOW if the controller is 6 button.
        move.b  #%00000011,d2                                  ; d2 =         00000011
        and.b   d1,d2                                          ; d2 = zero if six button controller present, else three button.
        bne     controllers_read_three_button_found
        and.w   #%0111111111111111,d0                          ; d0 = 00111111SACBRLDU 
controllers_read_three_button_found:                           ; d0 = 6P111111SACBRLDU  
        
        move.b  #$40,(a0)                                      ; Make TH high. (Cycle state 5.)
        nop                                                    ; Allow the controller to respond.
        nop
        move.b  (a0),d1                                        ; d1 =         xxxxMXYZ
        lsl.w   #8,d1                                          ; d1 = xxxxMXYZxxxxxxxx
        or.w    #%1111000011111111,d1                          ; d1 = 1111MXYZ11111111
        and.w   d1,d0                                          ; d0 = 6P11MXYZSACBRLDU        
        
        move.b  #$00,(a0)                                      ; Make TH low.  (Cycle state 6.)
        nop                                                    ; Allow the controller to respond.
        nop                                                    ; Let the controller idle to reset the counter. Leave TH low.
        
controllers_read_done:
        move.w  d0,(a1)

        lea     MM_CONTROLLER_DATA_SOURCE_STRIDE(a0),a0        ; Go to the next data source memory mapped register location
        lea     PAD_STATUS_STRUCT_STRIDE(a1),a1                ; Go to the next pad status struct
                                                                   
        dbf     d3,controllers_read_next_controller

        rts

; Make the inter-frame comparison values based on the current and previous states of the controllers.
; Destroys all data registers and a0.
controllers_derive:
        lea     pad_1_curr,a0                           ; Use a0 as reading and writing cursor.

        move.w  #1,d5                                   ; Loop through two controllers.
        
controllers_derive_next_controller:
        move.w  (a0)+,d0                                ; Load current into d0
        move.w  (a0)+,d1                                ; Load previous into d1

        move.w  d0,d2
        not.w   d2                                      ; d2 = !curr
        move.w  d1,d3
        not.w   d3                                      ; d3 = !prev
        ; Remember that the input values are active LOW, so our derived values are active LOW.
        
        move.w  d0,d4
        or.w    d3,d4                                   ; d4 =  curr | !prev = PRESSED
        move.w  d4,(a0)+

        move.w  d0,d4
        or.w    d1,d4                                   ; d4 =  curr |  prev = HELD
        move.w  d4,(a0)+

        move.w  d2,d4
        or.w    d1,d4                                   ; d4 = !curr |  prev = RELEASED
        move.w  d4,(a0)+

        move.w  d2,d4
        or.w    d3,d4                                   ; d4 = !curr | !prev = IDLE
        move.w  d4,(a0)+

        dbf     d5,controllers_derive_next_controller

        rts

GAME_RACCOON_install_game_at_directory_cache_record_index_d0:
        TIME_write 
GAME_RACCOON_install_game_at_directory_cache_record_index_d0_loop_until_deisolated:
        GAME_RACCOON_isolation_test
        bne     GAME_RACCOON_install_game_at_directory_cache_record_index_d0_loop_until_deisolated  
        rts 
        
GAME_RACCOON_game_partition_boot_sequence:
        move.w  #GAME_RACCOON_TIME_WRITE_MAGIC_NUMBER_BOOT_CURRENT_GAME_PARTITION,d0
        TIME_write
GAME_RACCOON_game_partition_boot_sequence_loop_until_deisolated:
        GAME_RACCOON_isolation_test
        bne     GAME_RACCOON_game_partition_boot_sequence_loop_until_deisolated
GAME_RACCOON_boot_current_game_partition:        ; Alright, we've switched to the game partition!
        move.l  (MM_STACK_POSITION),a7           ; Load the stack pointer starting value from the cart.
        move.l  (MM_ENTRY_POINT),a0              ; And jump to the entry point given in the cart image header!
        jmp     (a0)                             ; I wonder if any games will detect that the RAM and registers aren't clear...

        
        include "routines_maths.68k.asm"

        align   2                
        
; Variables are present in RAM beyond the installed RAM-resident block.
; We can use all the ram from asterisk up to however large we like
; as long as we don't crash into the stack.
; The assembler is smart enough to convert these labels into
; PC-relative addressing for shorter LEA instructions and so on.
; The only thing we need to do is make sure that all of these labels
; are consistent and monotonically increasing. 
; We CAN'T rely on align 2 to help us since we're not affecting the
; program counter.
; So we'll use things like aligned_longword_address = ((current_address + 3) & ~3)

variables_ram_begin = *

; Stores the current used rows of the following list:
directory_cache_items_list_status_struct_base                          = variables_ram_begin

directory_cache_items_list_status_struct_current_index                 = directory_cache_items_list_status_struct_base+4
directory_cache_items_list_status_struct_current_length                = directory_cache_items_list_status_struct_current_index+4
directory_cache_items_list_status_struct_no_directories                = directory_cache_items_list_status_struct_current_length+4
directory_cache_items_list_status_struct_no_files                      = directory_cache_items_list_status_struct_no_directories+4
directory_cache_items_list_status_struct_parent_index                  = directory_cache_items_list_status_struct_no_files+4
directory_cache_items_list_status_struct_no_directories_inc_parent_row = directory_cache_items_list_status_struct_parent_index+4

directory_cache_items_list_status_struct_end                           = directory_cache_items_list_status_struct_no_directories_inc_parent_row+4

; The directory cache items list contains the items can be shown on
; any single page in the interface.
; They are longword sized, since they are pointers (to directory entries ROM).
DIRECTORY_CACHE_ITEMS_LIST_SIZE_IN_ITEMS = 1024
directory_cache_items_list_begin = directory_cache_items_list_status_struct_end
directory_cache_items_list_end   = directory_cache_items_list_begin+(4*DIRECTORY_CACHE_ITEMS_LIST_SIZE_IN_ITEMS)

menu_state_cursor_location        = directory_cache_items_list_end+4
menu_state_scroll_top_row_shown   = menu_state_cursor_location+4

pad_state_base = menu_state_scroll_top_row_shown+4

; CONTROLLERS STATE.
pad_1_curr                = pad_state_base+(2*0)
pad_1_prev                = pad_state_base+(2*1)

pad_1_pressed             = pad_state_base+(2*2)       ; Pressed  =  curr & !prev
pad_1_held                = pad_state_base+(2*3)       ; Held     =  curr &  prev
pad_1_released            = pad_state_base+(2*4)       ; Released = !curr &  prev
pad_1_idle                = pad_state_base+(2*5)       ; Idle     = !curr & !prev

pad_2_curr                = pad_state_base+(2*6)
pad_2_prev                = pad_state_base+(2*7)

pad_2_pressed             = pad_state_base+(2*8)       ; Pressed  =  curr & !prev
pad_2_held                = pad_state_base+(2*9)       ; Held     =  curr &  prev
pad_2_released            = pad_state_base+(2*10)      ; Released = !curr &  prev
pad_2_idle                = pad_state_base+(2*11)      ; Idle     = !curr & !prev

PAD_STATUS_STRUCT_STRIDE = pad_2_curr-pad_1_curr

pad_state_end             = pad_state_base+(2*12)
; CONTROLLERS STATE END..